// Copyright 2007-2010 Baptiste Lepilleur and The JsonCpp Authors
// Distributed under MIT license, or public domain if desired and
// recognized in your jurisdiction.
// See file LICENSE for detail or copy at http://jsoncpp.sourceforge.net/LICENSE

#ifndef JSON_CONFIG_H_INCLUDED
#define JSON_CONFIG_H_INCLUDED
#include <cstddef>
#include <cstdint>
#include <istream>
#include <memory>
#include <ostream>
#include <sstream>
#include <string>
#include <type_traits>

// If non-zero, the library uses exceptions to report bad input instead of C
// assertion macros. The default is to use exceptions.
#ifndef JSON_USE_EXCEPTION
#	define JSON_USE_EXCEPTION 0
#endif

// Export macros for DLL visibility
#if defined(JSON_DLL_BUILD)
#	if defined(_MSC_VER) || defined(__MINGW32__)
#		define JSON_API __declspec(dllexport)
#		define JSONCPP_DISABLE_DLL_INTERFACE_WARNING
#	elif defined(__GNUC__) || defined(__clang__)
#		define JSON_API __attribute__((visibility("default")))
#	endif // if defined(_MSC_VER)

#elif defined(JSON_DLL)
#	if defined(_MSC_VER) || defined(__MINGW32__)
#		define JSON_API __declspec(dllimport)
#		define JSONCPP_DISABLE_DLL_INTERFACE_WARNING
#	endif // if defined(_MSC_VER)
#endif // ifdef JSON_DLL_BUILD

#if !defined(JSON_API)
#	define JSON_API
#endif

#ifndef JSONCPP_CPLUSPLUS
#	if defined(_MSVC_LANG) && !defined(__clang__)
#		define JSONCPP_CPLUSPLUS (_MSC_VER == 1900 ? 201103L : _MSVC_LANG)
#	else
#		define JSONCPP_CPLUSPLUS __cplusplus
#	endif
#endif

#if defined(_MSC_VER) && _MSC_VER < 1800
#error                                                                         \
	"ERROR:  Visual Studio 12 (2013) with _MSC_VER=1800 is the oldest supported compiler with sufficient C++11 capabilities"
#endif

#if defined(_MSC_VER) && _MSC_VER < 1900
// As recommended at
// https://stackoverflow.com/questions/2915672/snprintf-and-visual-studio-2010
extern JSON_API int msvc_pre1900_c99_snprintf(char* outBuf, size_t size,
											  const char* format, ...);
#	define jsoncpp_snprintf msvc_pre1900_c99_snprintf
#else
#	define jsoncpp_snprintf std::snprintf
#endif

// If JSON_NO_INT64 is defined, then Json only support C++ "int" type for
// integer
// Storages, and 64 bits integer support is disabled.
// #define JSON_NO_INT64 1

// JSONCPP_OVERRIDE is maintained for backwards compatibility of external tools.
// C++11 should be used directly in JSONCPP.
#define JSONCPP_OVERRIDE override

#ifdef __clang__
#	if __has_extension(attribute_deprecated_with_message)
#		define JSONCPP_DEPRECATED(message) __attribute__((deprecated(message)))
#	endif
#elif defined(__GNUC__) // not clang (gcc comes later since clang emulates gcc)
#	if (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 5))
#		define JSONCPP_DEPRECATED(message) __attribute__((deprecated(message)))
#	elif (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 1))
#		define JSONCPP_DEPRECATED(message) __attribute__((__deprecated__))
#	endif                  // GNUC version
#elif defined(_MSC_VER) // MSVC (after clang because clang on Windows emulates
						// MSVC)
#	define JSONCPP_DEPRECATED(message) __declspec(deprecated(message))
#endif // __clang__ || __GNUC__ || _MSC_VER

#if !defined(JSONCPP_DEPRECATED)
#	define JSONCPP_DEPRECATED(message)
#endif // if !defined(JSONCPP_DEPRECATED)

#if defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 6))
#	define JSON_USE_INT64_DOUBLE_CONVERSION 1
#endif

#include "allocator.h"
#include "version.h"

namespace Json
{
	using Int = int;
	using UInt = unsigned int;
#if defined(JSON_NO_INT64)
	using LargestInt = int;
	using LargestUInt = unsigned int;
#	undef JSON_HAS_INT64
#else                 // if defined(JSON_NO_INT64)
	// For Microsoft Visual use specific types as long long is not supported
#	if defined(_MSC_VER) // Microsoft Visual Studio
	using Int64 = __int64;
	using UInt64 = unsigned __int64;
#	else                 // if defined(_MSC_VER) // Other platforms, use long long
	using Int64 = int64_t;
	using UInt64 = uint64_t;
#	endif                // if defined(_MSC_VER)
	using LargestInt = Int64;
	using LargestUInt = UInt64;
#	define JSON_HAS_INT64
#endif // if defined(JSON_NO_INT64)

	template <typename T>
	using Allocator =
		typename std::conditional<JSONCPP_USE_SECURE_MEMORY, SecureAllocator<T>,
		std::allocator<T>>::type;
	using StringContainer = std::basic_string<char, std::char_traits<char>, Allocator<char>>;
	using IStringStream =
		std::basic_istringstream<StringContainer::value_type, StringContainer::traits_type,
		StringContainer::allocator_type>;
	using OStringStream =
		std::basic_ostringstream<StringContainer::value_type, StringContainer::traits_type,
		StringContainer::allocator_type>;
	using IStream = std::istream;
	using OStream = std::ostream;
} // namespace Json

// Legacy names (formerly macros).
using JSONCPP_STRING = Json::StringContainer;
using JSONCPP_ISTRINGSTREAM = Json::IStringStream;
using JSONCPP_OSTRINGSTREAM = Json::OStringStream;
using JSONCPP_ISTREAM = Json::IStream;
using JSONCPP_OSTREAM = Json::OStream;

#endif // JSON_CONFIG_H_INCLUDED
