//#pragma once
//
//#include <webp/decode.h>
//#include "ITextureLoader.h"
//
//namespace nCine {
//
///// WebP texture loader
//class TextureLoaderWebP : public ITextureLoader
//{
//  public:
//	explicit TextureLoaderWebP(std::unique_ptr<IFile> fileHandle);
//};
//
//}
//
