// Copyright © 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016,
//             2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024
//           Vladimír Vondruš <mosra@centrum.cz> and contributors
// Copyright © 2020-2024 Dan R.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.

#pragma once

/** @file
	@brief Forward declarations for the @ref Death::Containers namespace
*/

#include "../Common.h"

#include <type_traits>

#ifndef DOXYGEN_GENERATING_OUTPUT

namespace Death { namespace Containers {
//###==##====#=====--==~--~=~- --- -- -  -  -   -

	template<class T, class = void(*)(T*, std::size_t)> class Array;
	template<class> class ArrayView;
	template<std::size_t, class> class StaticArray;
	template<std::size_t, class> class StaticArrayView;

	template<class> class ComPtr;
	template<class, class> class Pair;
	template<class> class Pointer;
	template<class> class Reference;

	class FunctionData;
	template<class> class Function;

	enum class StringViewFlags : std::size_t;
	class String;
	template<class> class BasicStringView;
	typedef BasicStringView<const char> StringView;
	typedef BasicStringView<char> MutableStringView;

	class StringIterable;
	class StringIterableIterator;

	class DateTime;
	class TimeSpan;

	namespace Implementation
	{
		template<class, class, class = void> struct ArrayViewConverter;
		template<std::size_t, class, class, class = void> struct StaticArrayViewConverter;
		template<class, class = void> struct ErasedStaticArrayViewConverter;
	}

}}

#endif
